/*----------------------------------------------------------------
  Raven Library Source Code
  Copyright (c) 2008-2023 the Raven Development Team
  ----------------------------------------------------------------*/
#ifndef UNITTESTING_H
#define UNITTESTING_H

#include "RavenInclude.h"

void DateTest();
void OpticalAirMassTest();
void ClearSkyTest();
void ShortwaveTest();
void ShortwaveGenerator();
void JulianConvertTest();
void SmartIntervalTest();
void GammaTest();
void AddTimeTest( );
void BarycentricWeights();
void RavenUnitTesting(const optStruct &Options);
void TestInversion();
void TestEnthalpyTempConvert();
void TestConvectionSolution();
void ADRCumDistTest();
void TestGammaSampling();
void TestWetBulbTemps();
#endif
